from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from ftw.mail.interfaces import IEmailAddress
from plone.app.content.browser.folderfactories import _allowedTypes
from plone.app.layout.viewlets import ViewletBase
from zope.component import getMultiAdapter


class MailIn(ViewletBase):

    index = ViewPageTemplateFile('mail_templates/mail_in.pt')

    def available(self):
        """Only show viewlet if ftw.mail.mail is addable in current context
        """

        portal_types = getToolByName(self.context, 'portal_types')
        fti = portal_types.get(self.context.portal_type)
        return 'ftw.mail.mail' in fti.allowed_content_types

    def email(self):
        email = IEmailAddress(self.request)
        return email.get_email_for_object(self.context)



from ftw.builder import Builder
from ftw.builder import create
from ftw.mail.testing import FTW_MAIL_FUNCTIONAL_TESTING
from ftw.mail.tests.test_mail_view import mail_asset
from ftw.testbrowser import browsing
from StringIO import StringIO
from unittest2 import TestCase
from zipfile import ZipFile


class TestMailView(TestCase):
    layer = FTW_MAIL_FUNCTIONAL_TESTING

    def setUp(self):
        self.mail = create(Builder('mail').with_message(mail_asset('latin1')))

    @browsing
    def test_zipexport_works_on_mail(self, browser):
        browser.login().visit(self.mail, view='zip_export')
        self.assertEquals('application/zip', browser.headers['Content-Type'])
        self.assertEquals("attachment; filename*=utf-8''Die%20B%C3%BCrgschaft.zip",
                          browser.headers['Content-Disposition'])

        zipfile = ZipFile(StringIO(browser.contents))
        self.assertEquals(['message.eml'], zipfile.namelist())
        self.assertMultiLineEqual(mail_asset('latin1').read(),
                                  zipfile.read('message.eml'))



from ftw.upgrade import UpgradeStep


class IndexMailDate(UpgradeStep):
    """Update Date field in mail cache to new format
    and index mail date in the Date index field
    """

    def __call__(self):
        query = {'portal_type': 'ftw.mail.mail'}
        msg = 'Update E-mail cache.'
        for mail in self.objects(query, msg):
            if hasattr(mail, "_header_cache") and "Date" in mail._header_cache:
                del mail._header_cache["Date"]
        self.catalog_reindex_objects(query, idxs=['Date'])



from ftw.upgrade import UpgradeStep


class RebuildMailCacheForSignedMultipart(UpgradeStep):
    """Rebuild mail attachment info cache for signed/multipart.
    """

    deferrable = True

    def __call__(self):
        query = {'portal_type': 'ftw.mail.mail'}
        for mail in self.objects(query, 'Rebuild signed/multipart attachment'):
            self.maybe_rebuild_attachment_infos(mail)

    def maybe_rebuild_attachment_infos(self, mail):
        for info in mail.attachment_infos:
            if info.get('content-type') == 'multipart/signed':
                mail._update_attachment_infos()
                return



from Acquisition import aq_inner
from collective import dexteritytextindexer
from DateTime import DateTime
from email.MIMEText import MIMEText
from ftw.mail import _
from ftw.mail import utils
from persistent.mapping import PersistentMapping
from plone import api
from plone.dexterity.content import Item
from plone.directives import form
from plone.memoize import instance
from plone.namedfile import field
from plone.rfc822.interfaces import IPrimaryField
from premailer import transform as premailer_transform
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.deprecation import deprecate
from zope.interface import alsoProvides
from zope.interface import implements
import email


class IMail(form.Schema):
    """An e-mail message.
    """

    form.primary('message')
    message = field.NamedBlobFile(
        title=_(u"label_raw_message", default=u"Raw Message"),
        description=_(u"help_raw_message", default=u""),
        required=False,
    )


# necessary for dexterity 1 installations
alsoProvides(IMail['message'], IPrimaryField)


class Mail(Item):
    """An e-mail message."""

    @property
    def title(self):
        default_title = _(u'no_subject', default=u'[No Subject]')
        return getattr(self, '_title', default_title)

    @title.setter
    def title(self, value):
        """Since the title of an e-Mail can't be changed (it's always
        what the subject of the message contains), this is a dummy setter.
        It is still needed though, because otherwise plone.dexterity.utils
        fails in createContent() when trying to set attributes on the newly
        created content object.
        """
        pass

    def _update_title_from_message_subject(self):
        subject = utils.get_header(self.msg, 'Subject')
        if subject:
            # long headers may contain line breaks with tabs.
            # replace these by a space.
            subject = subject.replace('\n\t', ' ')
            self._title = subject.decode('utf8')

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, message):
        self._message = message
        self._update_title_from_message_subject()
        self._update_attachment_infos()
        self._reset_header_cache()

    @property
    def msg(self):
        """ returns an email.Message instance
        """
        if self.message is not None:
            data = self.message.data
            temp_msg = email.message_from_string(data)
            if temp_msg.get('Subject') and '\n\t' in temp_msg['Subject']:
                # It's a long subject header than has been separated by
                # line break and tab - fix it
                fixed_subject = temp_msg['Subject'].replace('\n\t', ' ')
                data = data.replace(temp_msg['Subject'], fixed_subject)
            return email.message_from_string(data)
        return MIMEText('')

    @property
    def attachment_infos(self):
        """A tuple of all attachment, each containing a dict of informations
        for the attachment.

        Example:
        {'position': 4,
        'size': 137588,
        'content-type': 'image/jpg',
        'filename': '1703693_0412c29a4f.jpg'}
        """
        infos = getattr(self, '_attachment_infos', ())
        # Make a copy of each info dict, so that users cannot modify
        # our persistent cache.
        return tuple(map(dict, infos))

    def _update_attachment_infos(self):
        self._attachment_infos = tuple(utils.get_attachments(self.msg))

    def get_header(self, name, isdate=False):
        """Returns a header value from the mail message.
        If it is a date, it returns a DateTime.
        This method caches the retrieved values.
        """

        if getattr(self, '_header_cache', None) is None:
            self._reset_header_cache()

        if name not in self._header_cache:
            if isdate:
                ts = utils.get_date_header(self.msg, name)
                if ts is not None:
                    self._header_cache[name] = DateTime(ts)
                else:
                    self._header_cache[name] = ""
            else:
                self._header_cache[name] = utils.get_header(self.msg, name)

        return self._header_cache[name]

    def _reset_header_cache(self):
        self._header_cache = PersistentMapping()


# SearchableText
class SearchableTextExtender(object):
    """This DynamicTextIndexExtender decodes the body of e-Mail messages
    and adds it to the searchableText.
    """

    implements(dexteritytextindexer.IDynamicTextIndexExtender)

    def __init__(self, context):
        self.context = context

    def __call__(self):
        transforms = getToolByName(self.context, 'portal_transforms')

        def convert_to_text(msg):
            if msg.is_multipart():
                result = map(convert_to_text, msg.get_payload())
                result = filter(None, result)
                return ' '.join(result)

            # Decode the Content-Transfer-Encoding (QP or base64)
            payload_data = msg.get_payload(decode=True)

            # Decode the actual content
            charset = msg.get_content_charset()
            if charset is not None:
                try:
                    payload_data = payload_data.decode(charset)
                except UnicodeDecodeError:
                    # Wrong charset declared. Try decoding with latin1
                    # as a last resort, and ignore any errors
                    payload_data = payload_data.decode('latin1', 'ignore')
            else:
                # No content charset declared - decode with utf-8
                # and hope for the best, ignoring any decoding errors
                payload_data = payload_data.decode('utf-8', 'ignore')

            # Finally encode it to UTF-8 (transforms require bytestrings)
            payload_data = payload_data.encode('utf-8')

            result = transforms.convertTo('text/plain',
                                          payload_data,
                                          mimetype=msg.get_content_type(),
                                          filename=msg.get_filename())

            if not result:
                return None

            result = result.getData().strip()

            if isinstance(result, unicode):
                return result.encode('utf-8')
            else:
                return result

        return convert_to_text(self.context.msg)


class View(BrowserView):
    """Default view for mail content type
    """

    template = ViewPageTemplateFile('mail_templates/view.pt')

    def __call__(self):
        return self.template()

    def get_header(self, name):
        return utils.get_header(self.msg(), name)

    def get_date_header(self, name):
        return DateTime(utils.get_date_header(self.msg(), name))

    @deprecate('use .html_safe_body() instead of .body()')
    def body(self):
        return self.html_safe_body()

    def rewrite_css_styles(self, body):
        """Rewrites CSS rules in <style /> tags to `style="..."` attributes.
        This is because <style /> tags will be removed by the `text/x-html-safe`
        transform, and leaving them in would mean that styles in a mail message could
        overstyle page elements outside the mail content area.
        """
        return premailer_transform(body.decode('utf-8'))

    def transfrom_safe_html(self, body):
        transformer = api.portal.get_tool('portal_transforms')
        return transformer.convertTo('text/x-html-safe', body).getData()

    def html_safe_body(self):
        """Converts the mail body to a html safe variant by using the
        following transforms:
         - the premailer css parser.
         - The `safe_html` PortalTranforms
        """
        context = aq_inner(self.context)
        parts = utils.get_body(self.msg(), context.absolute_url())
        if not parts:
            return ''

        result = []
        for body in parts:
            body = self.rewrite_css_styles(body)
            body = utils.fix_broken_meta_tags(body)
            body = utils.unwrap_html_body(body, 'mailBody-part')
            body = self.transfrom_safe_html(body)
            result.append(body)

        return '\n'.join(result)

    @instance.memoize
    def msg(self):
        context = aq_inner(self.context)
        return context.msg

    @instance.memoize
    def attachments(self):
        context = aq_inner(self.context)
        attachments = list(self.context.attachment_infos)
        mtr = getToolByName(context, 'mimetypes_registry')
        for attachment in attachments:
            icon = 'file_icon.gif'
            type_name = 'File'
            lookup = mtr.lookup(attachment['content-type'])
            if lookup:
                icon = lookup[0].icon_path
                type_name = lookup[0].name()
            attachment['icon'] = icon
            attachment['type-name'] = type_name
        return attachments



from ftw.upgrade import UpgradeStep


class ReindexSearchableText(UpgradeStep):

    def __call__(self):
        self.catalog_reindex_objects({'portal_type': 'ftw.mail.mail'},
                                     idxs=['SearchableText'])



from ftw.upgrade import UpgradeStep
from ZODB.POSException import ConflictError
import logging

logger = logging.getLogger('ftw.upgrade')


class FixMessageContentType(UpgradeStep):
    """Fix type of the NamedBlobFile's `contentType` attribute (unicode -> str)

    ftw.mail's `createMailInContainer` function previously set it to unicode,
    but it should be string.
    """

    def __call__(self):
        for obj in self.objects({'portal_type': 'ftw.mail.mail'},
                                'Fix message.contentType (unicode -> str)'):

            # ftw.mail's cached `message` property behaves in weird ways
            # sometimes, that's why this upgrade step takes an unususally
            # defensive approach

            # initialize attributes (lazy attribute loading)
            obj.Title()

            message = obj.message
            try:
                if isinstance(message.contentType, unicode):
                    message.contentType = message.contentType.encode('ascii')
            except ConflictError:
                raise
            except Exception, e:
                logger.warn("Updating object {0} failed: {1}".format(
                    obj.absolute_url(), str(e)))



# various utility methods for handling e-mail messages
from BeautifulSoup import BeautifulSoup
from DocumentTemplate.DT_Util import html_quote
from email.header import decode_header
from email.Utils import mktime_tz
from email.Utils import parsedate_tz
from ftw.mail import config
import re
import email


# a regular expression that matches src attributes of img tags containing a cid
IMG_SRC_RE = re.compile(r'<img[^>]*?src="cid:([^"]*)', re.IGNORECASE | re.DOTALL)
BODY_RE = re.compile(r'<body>(.*)</body>', re.IGNORECASE | re.DOTALL)
APPLE_PARTIAL_ENCODING_RE = re.compile(r'^"(.*=\?.*\?=.*)"( <.*>)$')
ENCODED_WORD_WITHOUT_LWSP = re.compile(r'(=\?.*?\?=)(\r\n)([ \t])')
ENCODED_WORD_WITHOUT_NEWLINES = re.compile(r'(=\?.*?\?[BQ]\?.*?)([\r\n])(.*?\?=)')
STICKY_ENCODED_WORDS = re.compile(r'(\?=)(=\?.*?\?[BQ]\?)')

# Used to fix broken meta tags that confuse TAL
# Largely copied from zope.pagetemplate.pagetemplatefile, adjusted for the
# missing semicolon between mimetype and charset declaration
broken_meta_pattern = re.compile(
    r'\s*<meta\s+http-equiv=["\']?Content-Type["\']?'
    r'\s+content=["\']?([^;]+)\s*charset=([^"\']+)["\']?\s*/?\s*>\s*',
    re.IGNORECASE)


def safe_decode_header(value):
    """Handles rfc 2047 encoded header with non-ascii characters.
    Always returns an utf-8 encoded string.
    """
    if value is None:
        return None

    if isinstance(value, unicode):
        value = value.encode('utf-8')

    new_value = []

    # When sending mails with Apple Mail, which have umlauts in
    # From-/To-/Cc-Headers, it encodes the name of the person but it
    # wraps it into quotes.
    # The python email module does not support this wrapping, so we
    # remove the quotes in this situation.
    # Example:
    # From: "=?iso-8859-1?Q?Boss_H=FCgo?=" <hugo@boss.com>
    apple_partial_encoding = APPLE_PARTIAL_ENCODING_RE.match(value)
    if apple_partial_encoding:
        value = ''.join(apple_partial_encoding.groups())

    # Fix up RFC 2047 encoded words separated by 'CRLF LWSP' (which is fine
    # according to the RFC) by replacing the CRLF with a SPACE so
    # decode_header can parse them correctly.
    # This works around a bug in decode_header that has been fixed in 3.3.
    # See http://bugs.python.org/issue4491 and its duplicate.
    # Example:
    # From: =?utf-8?Q?B=C3=A4rengraben?=\r\n <from@example.org>
    value = re.sub(ENCODED_WORD_WITHOUT_LWSP, '\\1 \\3', value)

    # Despite being allowed by RFC 1342, newlines inside encoded words will
    # break the python 2.7 decode_header.
    # https://github.com/python/cpython/blob/2.7/Lib/email/header.py#L78
    # This workaround removes newlines only from valid encoded words.
    value = re.sub(ENCODED_WORD_WITHOUT_NEWLINES, '\\1 \\3', value)

    # Normally the encoded words are separated by a space. But we have
    # seen cases in the wild where two encoded words "sticked" together, i.e.
    # were not separated by a space. The prefix of the encoded word
    # (=?charset?encoding?) then remained in the header which does not look nice
    # for humans.
    # Example: =?utf-8?Q?some?==?utf-8?Q?thing?=
    # Should be: =?utf-8?Q?some?= =?utf-8?Q?thing?=
    # We fix this by separating those "sticky" encoded words with a space.
    # The regex used to do this is based on the following idea: if the end
    # of an encoded word (?=) is followed by the beginning of an encoded word
    # (=?charset?encoding?) without a whitespace between them, we need to
    # separate them.
    value = re.sub(STICKY_ENCODED_WORDS, '\\1 \\2', value)

    for data, charset in decode_header(value):
        if charset is not None and charset not in ('utf-8', 'utf8'):
            data = data.decode(charset).encode('utf-8')
        elif charset is None:
            data = safe_utf8(data)
        new_value.append(data and data or '')

    return ' '.join(new_value).rstrip()


def get_header(msg, name):
    """Returns the value of the named header field of a mail message.
    Handles rfc 2047 encoded header with non-ascii characters.
    Always returns an utf-8 encoded string.
    """
    value = ''
    if msg is not None and name in msg:
        value = safe_decode_header(msg.get(name))
    return value


def get_date_header(msg, name):
    """Returns an UTC timestamp from a header field containing a date.
    Compensates for the timezone difference if the header contains
    timezone information.
    """
    value = get_header(msg, name)
    ts = None
    try:
        ts = mktime_tz(parsedate_tz(value))
    except TypeError:
        pass
    return ts


def get_payload(msg):
    """Get the decoded message payload as utf-8 string."""
    # get encoding for the msg object, use utf-8 as fallback encoding
    # for msg objects without a correct charset information
    encoding = msg.get_content_charset('utf-8')
    payload = msg.get_payload(decode=1)
    try:
        payload = payload.decode(encoding)
    except UnicodeDecodeError:
        payload = payload.decode('iso-8859-1', 'replace')
    payload = payload.encode('utf-8')
    return payload


def get_body(msg, url_prefix=''):
    """Returns the mail body as HTML string. All text parts of a multipart
    message are returned.
    """
    result = []
    for part in get_text_payloads(msg):
        result.append(adjust_image_tags(part, msg, url_prefix))
    return result


def get_part_size(content_type, part):
    if content_type == 'message/rfc822':
        return len(part.as_string())

    payload = part.get_payload(decode=1) or part.get_payload()
    if isinstance(payload, str):
        return len(payload)

    if not payload:
        return 0

    return sum(get_part_size(content_type, part) for part in payload)


def parse_part(part):
    content_type = part.get_content_type()
    filename = get_filename(part)
    size = get_part_size(content_type, part)
    return content_type, filename, size


def get_attachments(msg):
    """Returns a list describing the attachements. Only attachments with
    a filename are returned.
    """
    attachments = []

    if msg is not None and msg.is_multipart():
        for position, part in enumerate(walk(msg)):
            content_type, filename, size = parse_part(part)

            if filename is None:
                continue

            attachments.append({
                'filename': filename,
                'content-type': content_type,
                'size': size,
                'position': position,
            })

    return attachments


def remove_attachments(msg, positions):
    """Remove all attachments which have position listed in `positions`
    from the email Message `msg`.

    Returns the same email Message object without attachments.
    If `clone` is True, the message will be copied first.
    """
    if msg is None:
        raise ValueError('Cannot delete attachments from email message None')

    if 0 in positions:
        raise ValueError('Cannot delete the message itself (position 0)')

    if not positions:
        # no attachments should be removed - so just skip
        return msg

    if not msg.is_multipart():
        raise ValueError('Email message is not multipart - there are '
                         'no attachments.')

    # get the parts to delete filtering the position
    parts_to_delete = [part for pos, part in enumerate(walk(msg)) if pos in positions]

    if len(positions) != len(parts_to_delete):
        raise ValueError('One or more attachments could not be found.')

    def _recursive_remove_parts(msg):
        """Recursive function for removing payloads (attachments).
        """
        if not msg.is_multipart():
            return msg

        new_payload = []
        for part in msg.get_payload():
            if part not in parts_to_delete:
                new_payload.append(_recursive_remove_parts(part))
        msg.set_payload(new_payload)
        return msg

    # do it
    return _recursive_remove_parts(msg)


def get_text_payloads(msg):
    """Go recursivly through the message parts and return a list of all
    text parts in HTML format.
    """
    if msg is None:
        return []
    parts = []
    if msg.is_multipart():
        msgs = msg.get_payload()
        # we only want the best alternative
        if msg.get_content_type() == 'multipart/alternative':
            alternatives = []
            for part in msgs:
                alternatives.append(part.get_content_type())
            best_pos = get_best_alternative(alternatives)
            parts += get_text_payloads(msgs[best_pos])
        # we go through all parts
        else:
            for part in msgs:
                parts += get_text_payloads(part)
    # not a multipart message
    else:
        payload = get_payload(msg)
        if payload:
            if msg.get_content_type() == 'text/html':
                parts.append(payload)
            elif msg.get_content_type() == 'text/plain':
                parts.append(text2html(payload))
    return parts


def adjust_image_tags(html, msg, url_prefix):
    """Adjust image tags of the given HTML string which reference an image by
    Content-Id. The src attribute is set to the attachment of the given
    message.

    """
    matches = IMG_SRC_RE.findall(html)
    for m in matches:
        pos = get_position_for_cid(msg, m)
        if pos:
            download_url = '%s/get_attachment?position=%s' % (url_prefix, pos)
            html = html.replace('cid:' + m, download_url)
    return html


def get_position_for_cid(msg, cid):
    """Return the position of the message part with the given Content-Id."""
    position = -1
    cid = '<' + cid + '>'
    for part in walk(msg):
        position += 1
        if part.get('Content-Id', None) == cid:
            return position
    return None


def get_filename(msg):
    """Get the filename of a message (part)
    """
    filename = msg.get_filename(None)

    if filename is None:
        filename = msg.get_param('Name', None)

    # if the value is already decoded or another tuple
    # we just take the value and use the decode_header function
    if isinstance(filename, tuple):
        if filename[0] is not None and filename[0] not in ('utf-8', 'utf8'):
            filename = filename[2].decode(filename[0]).encode('utf-8')
        else:
            filename = safe_utf8(filename[2])

    else:
        filename = safe_decode_header(filename)

    return filename


def get_best_alternative(alternatives):
    """Get the index of the most preferred alternative in the given list of
    alternatives.
    """
    for content_type in config.PREFERRED_MULTIPART_ALTERNATIVES:
        if content_type in alternatives:
            return alternatives.index(content_type)
    return 0


def text2html(text):
    """Replaces all line breaks with a br tag, and wraps it in a p tag.
    """
    return '<p>%s</p>' % html_quote(text.strip()).replace('\n', '<br />')


def unwrap_html_body(html, css_class=None):
    """Return the content of the body tag for inline display in another
    html document.
    """
    soup = BeautifulSoup(html, fromEncoding='utf8')
    if soup.body:
        soup = soup.body
    body_soup = BeautifulSoup('<div>%s</div>' % soup.renderContents(), fromEncoding='utf8')
    if css_class:
        body_soup.div['class'] = css_class
    body_style = soup.get('style')
    if body_style:
        body_soup.div['style'] = body_style
    return body_soup.renderContents()


def fix_broken_meta_tags(html):
    """Fix broken <meta /> tags in HTML MIME parts.

    text/html MIME parts of a multipart message may include <meta /> tags that
    declare a charset that differs from the encoding that is used for the page
    template that part is inserted into.

    zope.pagetemplate attempts to handle meta tags when inserting markup
    with 'structure', by re-encoding the inserted markup and adjusting the
    rewriting the meta tag's charset.

    If however the <meta /> tag is broken and doesn't match zope.pt's regex,
    the re-encoding doesn't take place and TAL / zope.pt chokes on the
    resulting mojibake, omitting parts of the document.

    This addresses one specific case seen in the wild in a message created
    with Apple Mail (2.2104), where the semicolon (;) is missing between the
    MIME type and the charset declaration. Example:

    <meta http-equiv="Content-Type" content="text/html charset=us-ascii">
    """
    match = broken_meta_pattern.search(html)

    if match:
        mimetype, charset = [g.strip() for g in match.groups()]
        fixed_tag = '\n<meta http-equiv="Content-Type" '\
                    'content="%s; charset=%s">\n' % (mimetype, charset)
        return re.sub(broken_meta_pattern, fixed_tag, html)

    return html


def unwrap_attached_msg(msg):
    """If a msg contains an attachmed message return the attached one."""
    if msg.is_multipart():
        for part in walk(msg):
            content_type = part.get_content_type()
            if content_type == 'message/rfc822':
                return part.get_payload(0)
    return msg


def safe_utf8(text):
    """Returns an utf-8 encoded version of the given string with unknown
    encoding.
    """
    encodings = ('utf8', 'iso-8859-1', 'iso-8859-15')
    for enc in encodings:
        try:
            text = text.decode(enc)
            return text.encode('utf8')
        except UnicodeDecodeError:
            pass
    return text.decode('ascii', 'replace').encode('utf8')


def walk(message):
    """Walk and recursively yield a message's parts in depth-first order.

    Also add special handling to yield parts of multipart/signed messages where
    one level needs to be unwrapped to correctly parse the message.
    """
    for part in message.walk():
        yield part
        content_type = part.get_content_type()
        if content_type == 'multipart/signed':
            payload = part.get_payload()
            if isinstance(payload, basestring):
                unwrapped = email.message_from_string(payload)
                for subpart in walk(unwrapped):
                    yield subpart



from ftw.builder import Builder
from ftw.builder import create
from ftw.mail.testing import FTW_MAIL_FUNCTIONAL_TESTING
from ftw.table import helper
from ftw.workspace.interfaces import IWorkspaceLayer
from plone.registry.interfaces import IRegistry
from unittest2 import TestCase
from zope.component import getUtility
from zope.interface import alsoProvides
import json
import os.path


def mail_asset(name, ext='txt'):
    tests_dir_path = os.path.dirname(__file__)
    filename = '.'.join((name, ext))
    return open(os.path.join(tests_dir_path, 'mails', filename))


class TestMailTab(TestCase):

    layer = FTW_MAIL_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestMailTab, self).setUp()

        self.portal = self.layer.get('portal')

    def test_mail_tab_available(self):
        alsoProvides(self.portal.REQUEST, IWorkspaceLayer)
        view = self.portal.restrictedTraverse('tabbedview_view-mails')

        self.assertTrue(view, 'Mail tab is not available')

    def test_mail_date_parsing(self):
        create(Builder('mail').with_message(mail_asset('latin1')))
        mail_row = self.get_mails_tab_data().get('rows')[0]
        self.assertEquals('01.01.1970 01:00', mail_row.get('Date'))

    def test_mail_invalid_date_results_in_creation_date(self):
        mail = create(Builder('mail').with_message(mail_asset('invalid_date')))
        mail_row = self.get_mails_tab_data().get('rows')[0]
        created = helper.readable_date_time_text(mail, mail.created())
        self.assertEquals(created, mail_row.get('Date'))

    def test_mail_date_parsing_with_time_zone(self):
        create(Builder('mail').with_message(mail_asset('time_zone_dates')))
        mail_row = self.get_mails_tab_data().get('rows')[0]
        self.assertEquals('28.08.2010 18:50', mail_row.get('Date'))

    def get_mails_tab_data(self):
        request = self.portal.REQUEST
        alsoProvides(request, IWorkspaceLayer)
        registry = getUtility(IRegistry)
        registry['ftw.tabbedview.interfaces.ITabbedView.extjs_enabled'] = True
        request.form['tableType'] = 'extjs'
        view = self.portal.restrictedTraverse('tabbedview_view-mails')
        return json.loads(view())



from DateTime import DateTime
from DateTime.interfaces import SyntaxError
from ftw.mail import _
from ftw.tabbedview.browser import listing
from ftw.table import helper
from ftw.table.helper import readable_author
from zope.i18n import translate
import re


def get_mail_header(field=None, isdate=False):

    def _helper(item, value):
        if field is None:
            return ''

        obj = item.getObject()

        if isdate:
            date = obj.get_header(field, isdate)
            return helper.readable_date_time_text(item, date)

        elif field == 'attachments':
            imgpath = '++resource++attachment.png'
            return '<img src="%s" alt="%s" /> %s' % (
                imgpath,
                translate(_(u'attachment_icon_alt_text',
                            default=u'Attachment'),
                          context=obj.REQUEST),
                len(obj.attachment_infos))

        else:
            return obj.get_header(field)

    return _helper


class MailsTab(listing.CatalogListingView):
    """Lists all mails
    """

    types = ('ftw.mail.mail',)

    sort_on = 'Date'
    sort_reverse = True
    show_selects = False
    show_menu = False

    columns = (
        {'column': 'Title',
         'sort_index': 'sortable_title',
         'column_title': _(u'label_mailtab_subject',
                           default=u'Subject'),
         'transform': helper.linked},

        {'column': 'From',
         'column_title': _(u'label_mailstab_from',
                           default=u'From'),
         'transform': get_mail_header(field='From'),
         'sortable': False},

        {'column': 'To',
         'column_title': _(u'label_mailstab_to',
                           default=u'To'),
         'transform': get_mail_header(field='To'),
         'sortable': False},

        {'column': 'Date',
         'column_title': _(u'label_mailstab_date',
                           default=u'Date Received'),
         'transform': helper.readable_date_time_text,
         'sortable': True},

        {'column': 'Attachments',
         'column_title': _(u'label_mailstab_attachments',
                           default=u'Attachments'),
         'transform': get_mail_header(field='attachments'),
         'sortable': False},

        {'column': 'Creator',
         'column_title': _(u'label_mailstab_creator',
                           default=u'Creator'),
         'transform': readable_author},
    )



from AccessControl import getSecurityManager
from AccessControl import Unauthorized
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import setSecurityManager
from email.Utils import parseaddr
from ftw.mail import exceptions
from ftw.mail import utils
from ftw.mail.interfaces import ICreateMailInContainer
from ftw.mail.interfaces import IEmailAddress
from ftw.mail.interfaces import IInboundRequest
from ftw.mail.interfaces import IMailInbound
from ftw.mail.interfaces import IMailSettings
from plone.memoize import instance
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.interface import alsoProvides
from zope.interface import implements
from zope.interface import noLongerProvides
import email
import logging


log = logging.getLogger('ftw.mail.inbound')


class MailInbound(BrowserView):
    """ A view that handles inbound mail posted by mta2plone.py
    """

    implements(IMailInbound)

    def __call__(self):
        alsoProvides(self.request, IInboundRequest)
        result = self.render()
        noLongerProvides(self.request, IInboundRequest)
        return result

    def render(self):
        self.request.response.setHeader('Content-Type', 'text/plain')
        try:
            self.inbound()
            return '0:OK'
        except exceptions.MailInboundException, e:
            log.warn('Rejected inbound mail attempt: %r' % str(e))
            return str(e)

    def inbound(self):
        msg = self.msg()
        settings = getUtility(IRegistry).forInterface(IMailSettings)
        if settings.unwrap_mail:
            # if we find an attached mail, use this instead of the whole one
            msg = utils.unwrap_attached_msg(msg)

        user = self.get_user()
        sm = getSecurityManager()
        newSecurityManager(self.request, user)
        try:
            destination = self.get_destination()
            createMailInContainer(destination, msg.as_string())
        except Unauthorized:
            log.warn('User %r is not allowed to create mail '
                     'in %r' % (user, destination))
            raise exceptions.PermissionDenied(self.msg(), user)
        except ValueError:
            # ValueError can also be caused by non-intid destination address
            if 'destination' in locals():
                log.warn('Attempt to create mail in %r' % destination)
            raise exceptions.DisallowedSubobjectType(self.msg(), user)
        finally:
            setSecurityManager(sm)

    def get_user(self):
        sender_email = self.sender()
        if not sender_email:
            log.warn("No sender found in 'Resent-From' or 'From' headers")
            raise exceptions.NoSenderFound(self.msg())

        acl_users = getToolByName(self.context, 'acl_users')
        pas_search = getMultiAdapter((self.context, self.request),
                                     name='pas_search')
        users = pas_search.searchUsers(email=sender_email)
        if len(users) > 0:
            user = acl_users.getUserById(users[0].get('userid'))
            if not hasattr(user, 'aq_base'):
                user = user.__of__(acl_users)
            return user
        else:
            log.warn('Sender address: %r' % sender_email)
            log.warn('No users found in %r for that address.' % acl_users)
            raise exceptions.UnknownSender(self.msg())

    def get_destination(self):
        emailaddress = IEmailAddress(self.request)
        destination = emailaddress.get_object_for_email(self.recipient())

        if destination is None:
            log.warn('Destination not found: %r' % self.recipient())
            raise exceptions.DestinationDoesNotExist(self.recipient())
        return destination

    @instance.memoize
    def msg(self):
        """ create an email.Message instance from request.
        """
        msg_txt = self.request.get('mail', None)
        if msg_txt is None:
            raise exceptions.NoInput()
        try:
            msg = email.message_from_string(msg_txt)
        except TypeError:
            raise exceptions.InvalidMessage()
        return msg

    @instance.memoize
    def sender(self):
        """ get the email address of the sender
        """
        sender = utils.get_header(self.msg(), 'Resent-From')
        if not sender:
            sender = utils.get_header(self.msg(), 'From')
            log.info("Picked sender from 'From' header")
        else:
            log.info("Picked sender from 'Resent-From' header")
        (sender_name, sender_address) = parseaddr(sender)
        return sender_address

    @instance.memoize
    def recipient(self):
        """ get the email address of the recipient
        """
        # If recipient is supplied in request (by mta2plone), use that
        recipient = self.request.get('recipient')
        if not recipient:
            # Otherwise fall back to `Resent-To` or `To` headers
            recipient = utils.get_header(self.msg(), 'Resent-To')
            if not recipient:
                recipient = utils.get_header(self.msg(), 'To')
        (recipient_name, recipient_address) = parseaddr(recipient)
        return recipient_address


def createMailInContainer(container, message):
    """Add a mail object to a container.

    The new object, wrapped in its new acquisition context, is returned.
    """
    creator = getMultiAdapter((container, container.REQUEST),
                              ICreateMailInContainer)
    return creator.create_mail(message)



from pkg_resources import resource_string
import email


def load(filename):
    return resource_string('ftw.mail.tests.mails', filename)


def load_mail(filename):
    return email.message_from_string(load(filename))



# -*- coding: utf-8 -*-
from copy import deepcopy
from email.MIMEText import MIMEText
from ftw.mail import utils
from ftw.mail.tests import mails
import email
import unittest2


def pop_size(attachments):
    """Pop fs/os dependent size from attachment metadata dict.
    """
    for each in attachments:
        each.pop('size', None)
    return attachments


class TestUtils(unittest2.TestCase):

    maxDiff = None

    def setUp(self):
        # setup some test mails
        self.msg_empty = MIMEText('')
        self.msg_ascii = mails.load_mail(
            'ascii_7bit.txt')
        self.msg_latin1 = mails.load_mail(
            'latin1.txt')
        self.msg_utf8 = mails.load_mail(
            'utf8.txt')
        self.msg_attachment = mails.load_mail(
            'attachment.txt')
        self.msg_fwd_attachment = mails.load_mail(
            'fwd_attachment.txt')
        self.msg_nested_attachments = mails.load_mail(
            'nested_attachments.txt')
        self.nested_referenced_image_attachment = mails.load_mail(
            'nested_referenced_image_attachment.txt')
        self.msg_multiple_html_parts = mails.load_mail(
            'multiple_html_parts.txt')
        self.multipart_encoded_with_attachments = mails.load_mail(
            'multipart_encoded_with_attachments.txt')
        self.from_header_with_quotes = mails.load_mail(
            'from_header_with_quotes.txt')
        self.encoded_word_without_lwsp = mails.load_mail(
            'encoded_word_without_lwsp.txt')
        self.newline_in_header = mails.load_mail(
            'newline_in_header.txt')
        self.sticky_encoded_words_in_subject = mails.load_mail(
            'encoded_word_not_separated_by_lwsp.txt')

    def test_walk_signed(self):
        msg = mails.load_mail('signed.eml')
        parts = list(utils.walk(msg))

        self.assertEqual(7, len(parts))
        self.assertEqual(
            ['multipart/mixed', 'text/plain', 'multipart/signed',
             'multipart/signed', 'multipart/mixed', 'text/plain',
             'application/x-pkcs7-signature'],
            [each.get_content_type() for each in parts])

    def test_get_attachments_signed(self):
        msg = mails.load_mail('signed.eml')
        self.assertEquals(
            [{'position': 2,
            'content-type': 'multipart/signed',
            'filename': 'smime.p7m'},
            {'position': 6,
             'content-type':
             'application/x-pkcs7-signature',
             'filename': 'smime.p7s'}],
            pop_size(utils.get_attachments(msg)))

    def test_get_attachments_signed_with_attachments(self):
        msg = mails.load_mail('signed_with_attachments.eml')
        self.assertEquals(
           [{'position': 4,
            'content-type': 'multipart/signed',
            'filename': 'smime.p7m'},
           {'position': 8,
            'content-type': 'application/pdf',
            'filename': 'Testdatei.pdf'},
           {'position': 9,
            'content-type': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
            'filename': 'Testdatei.docx'},
           {'position': 10,
            'content-type': 'application/pkcs7-signature',
            'filename': 'smime.p7s'}],
           pop_size(utils.get_attachments(msg)))

    def test_walk_signed_with_attachments(self):
        msg = mails.load_mail('signed_with_attachments.eml')
        parts = list(utils.walk(msg))

        self.assertEqual(11, len(parts))
        self.assertEqual(
            ['multipart/mixed', 'multipart/alternative', 'text/plain',
             'application/rtf', 'multipart/signed', 'multipart/signed',
             'multipart/mixed', 'text/plain', 'application/pdf',
             'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
             'application/pkcs7-signature'],
            [each.get_content_type() for each in parts])

    def test_get_attachments_signed_nested_with_attachments(self):
        msg = mails.load_mail('signed_nested_with_attachments.eml')
        self.assertEquals(
            [{'position': 4,
              'content-type': 'multipart/signed',
              'filename': 'smime.p7m'},
             {'position': 12,
              'content-type': 'application/pdf',
              'filename': 'Testdatei.pdf'},
             {'position': 13,
              'content-type': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
              'filename': 'Testdatei.docx'},
             {'position': 14,
              'content-type': 'application/pkcs7-signature',
              'filename': 'smime.p7s'},
             {'position': 15,
              'content-type': 'application/pkcs7-signature',
              'filename': 'smime.p7s'}],
            pop_size(utils.get_attachments(msg)))

    def test_walk_signed_nested_with_attachments(self):
        msg = mails.load_mail('signed_nested_with_attachments.eml')
        parts = list(utils.walk(msg))

        self.assertEqual(16, len(parts))
        self.assertEqual(
            ['multipart/mixed', 'multipart/alternative', 'text/plain',
             'application/rtf', 'multipart/signed', 'multipart/signed',
             'multipart/mixed', 'text/plain', 'message/rfc822',
             'multipart/signed', 'multipart/mixed', 'text/plain',
             'application/pdf',
             'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
             'application/pkcs7-signature', 'application/pkcs7-signature'],
            [each.get_content_type() for each in parts])

    def test_walk_util_and_walk_yield_same_for_non_signed_multipart(self):
        self.assertEqual(
            list(self.msg_empty.walk()),
            list(utils.walk(self.msg_empty)))
        self.assertEqual(
            list(self.msg_ascii.walk()),
            list(utils.walk(self.msg_ascii)))
        self.assertEqual(
            list(self.msg_latin1.walk()),
            list(utils.walk(self.msg_latin1)))
        self.assertEqual(
            list(self.msg_utf8.walk()),
            list(utils.walk(self.msg_utf8)))
        self.assertEqual(
            list(self.msg_attachment.walk()),
            list(utils.walk(self.msg_attachment)))
        self.assertEqual(
            list(self.msg_fwd_attachment.walk()),
            list(utils.walk(self.msg_fwd_attachment)))
        self.assertEqual(
            list(self.msg_nested_attachments.walk()),
            list(utils.walk(self.msg_nested_attachments)))
        self.assertEqual(
            list(self.nested_referenced_image_attachment.walk()),
            list(utils.walk(self.nested_referenced_image_attachment)))
        self.assertEqual(
            list(self.msg_multiple_html_parts.walk()),
            list(utils.walk(self.msg_multiple_html_parts)))
        self.assertEqual(
            list(self.multipart_encoded_with_attachments.walk()),
            list(utils.walk(self.multipart_encoded_with_attachments)))
        self.assertEqual(
            list(self.from_header_with_quotes.walk()),
            list(utils.walk(self.from_header_with_quotes)))
        self.assertEqual(
            list(self.encoded_word_without_lwsp.walk()),
            list(utils.walk(self.encoded_word_without_lwsp)))
        self.assertEqual(
            list(self.newline_in_header.walk()),
            list(utils.walk(self.newline_in_header)))
        self.assertEqual(
            list(self.sticky_encoded_words_in_subject.walk()),
            list(utils.walk(self.sticky_encoded_words_in_subject)))

    def test_get_header(self):
        self.assertEquals('', utils.get_header(self.msg_empty, 'Subject'))
        self.assertEquals('Lorem Ipsum',
                          utils.get_header(self.msg_ascii, 'Subject'))
        self.assertEquals('Die B\xc3\xbcrgschaft',
                          utils.get_header(self.msg_latin1, 'Subject'))
        self.assertEquals('Friedrich H\xc3\xb6lderlin <to@example.org>',
                          utils.get_header(self.msg_latin1, 'To'))
        self.assertEquals('Die B\xc3\xbcrgschaft',
                          utils.get_header(self.msg_utf8, 'Subject'))
        self.assertEquals('Friedrich H\xc3\xb6lderlin <to@example.org>',
                          utils.get_header(self.msg_utf8, 'To'))

    def test_get_from_header_with_quotes(self):
        self.assertEquals(
            '"Mueller-Valser, Gabriela" <gabriela.mueller@example.org>',
            self.from_header_with_quotes['From'])

        self.assertEquals(
            '"Mueller-Valser, Gabriela" <gabriela.mueller@example.org>',
            utils.get_header(self.from_header_with_quotes, 'From'))

    def test_get_header_fixes_encoded_words_without_lwsp(self):
        self.assertEquals(
            'B\xc3\xa4rengraben <to@example.org>',
            utils.get_header(self.encoded_word_without_lwsp, 'To'))

        # Should not insert additional whitespace between encoded words
        self.assertEquals(
            'B\xc3\xa4rengrabenB\xc3\xa4rengraben <from@example.org>',
            utils.get_header(self.encoded_word_without_lwsp, 'From'))

    def test_get_header_fixes_encoded_words_with_newlines(self):
        self.assertEquals(
            'Email: QP B\xc3\xbcschelisheimat   / Aktennotiz vom 27.02.2017',
            utils.get_header(self.newline_in_header, 'Subject')
        )

    def test_safe_decode_header_fixes_encoded_words_without_lwsp_in_middle(self):
        header = 'Foo =?utf-8?Q?B=C3=A4rengraben?=\r\n <from@example.org>'
        self.assertEquals(
            'Foo B\xc3\xa4rengraben <from@example.org>',
            utils.safe_decode_header(header))

    def test_get_subject_header_with_sticky_encoded_words(self):
        self.assertEquals(
            'Bärengraben',
            utils.get_header(self.sticky_encoded_words_in_subject, 'Subject'))

    def test_get_date_header(self):
        # a date header
        msg_txt = 'Date: Thu, 01 Jan 1970 01:00:00 +0100'
        msg = email.message_from_string(msg_txt)
        self.assertEquals(0.0, utils.get_date_header(msg, 'Date'))
        # a date header with timezone name
        msg_txt = 'Date: Sat, 14 Feb 2009 00:31:30 +0100 (CET)'
        msg = email.message_from_string(msg_txt)
        self.assertEquals(1234567890.0, utils.get_date_header(msg, 'Date'))
        # an unparsable date header
        msg_txt = 'Date: at any time ...'
        msg = email.message_from_string(msg_txt)
        self.assertEqual(None, utils.get_date_header(msg, 'Date'))

    def test_get_payload(self):
        self.assertEquals('', utils.get_payload(self.msg_empty))
        self.assertEquals('Lorem ipsum', utils.get_payload(self.msg_ascii)[:11])
        self.assertEquals('Die B\xc3\xbcrgschaft', utils.get_payload(self.msg_latin1)[:15])
        self.assertEquals('Die B\xc3\xbcrgschaft', utils.get_payload(self.msg_utf8)[:15])

    def test_get_best_alternative(self):
        # prefer html over plain text
        alternatives = ['text/html', 'text/plain']
        self.assertEquals(0, utils.get_best_alternative(alternatives))

    def test_get_text_payloads(self):
        self.assertEquals([], utils.get_text_payloads(None))
        self.assertEquals(1, len(utils.get_text_payloads(self.msg_ascii)))

    def test_get_body(self):
        self.assertEquals(0, len(utils.get_body(self.msg_empty)))
        self.assertEquals('<p>Lorem ipsum', utils.get_body(self.msg_ascii)[0][:14])

    def test_get_body_returns_each_html_document_separately(self):
        parts = utils.get_body(self.msg_multiple_html_parts)
        self.assertEquals(2, len(parts), 'Expected two html parts.')
        self.assertIn('Hello', parts[0])
        self.assertIn('World', parts[1])

    def test_can_decode_encoded_multipart_attachments(self):
        expected_attachments = [{
            'content-type': 'message/delivery-status',
            'filename': 'ATT74209.txt',
            'position': 4,
            'size': 0,
            }]
        attachments = utils.get_attachments(self.multipart_encoded_with_attachments)
        self.assertEqual(expected_attachments, attachments)

    def test_get_filename(self):
        msg_txt = \
            """Content-Type: application/octet-stream;
  name="=?iso-8859-1?Q?Aperovorschl=E4ge_2010=2Epdf?="
Content-Transfer-Encoding: base64
Content-Description: =?iso-8859-1?Q?Aperovorschl=E4ge_2010=2Epdf?=
Content-Disposition: attachment;
  filename="=?iso-8859-1?Q?Aperovorschl=E4ge_2010=2Epdf?="
"""
        msg = email.message_from_string(msg_txt)
        # !!! seems to be a bug in email package
        self.assertEquals('Aperovorschläge 2010.pdf', utils.get_filename(msg))

        msg_txt = \
        """Content-Disposition: attachment;
	filename*=iso-8859-1''f%F6rmularz%FCgriffsber%E4chtigungen.doc
Content-Type: application/msword;
	name*=iso-8859-1''f%F6rmularz%FCgriffsber%E4chtigungen.doc
Content-Transfer-Encoding: base64
"""
        msg = email.message_from_string(msg_txt)
        self.assertEquals('f\xc3\xb6rmularz\xc3\xbcgriffsber\xc3\xa4chtigungen.doc', utils.get_filename(msg))

    def test_get_attachments(self):
        self.assertEquals([], utils.get_attachments(self.msg_ascii))
        self.assertEquals([{'position': 1,
                            'size': 7,
                            'content-type': 'text/plain',
                            'filename': 'Bücher.txt'}],
                          utils.get_attachments(self.msg_attachment))

    def test_nested_get_attachments(self):
        """A forwarded mail with attachments results in nested multipart
        payloads - this should also be handled by get_attachments method.
        """
        self.assertEquals([{'position': 4,
                            'size': 137588,
                            'content-type': 'image/jpg',
                            'filename': '1703693_0412c29a4f.jpg'},
                           {'position': 5,
                            'size': 223504,
                            'content-type': 'image/jpg',
                            'filename': '3512536451_e1310bf568.jpg'}],
                          utils.get_attachments(self.msg_nested_attachments))

    def test_remove_attachments(self):
        # we dont want to change the message itselve, so lets copy it
        msg = deepcopy(self.msg_attachment)
        self.assertNotEquals(msg, self.msg_attachment)

        # our message has one attachment
        self.assertEquals([{'position': 1,
                            'size': 7,
                            'content-type': 'text/plain',
                            'filename': 'Bücher.txt'}],
                          utils.get_attachments(msg))

        # lets remove the attachment
        new_msg = utils.remove_attachments(msg, (1,))

        # we get the same message back
        self.assertEquals(msg, new_msg)
        self.assertEquals([], utils.get_attachments(new_msg))

    def test_nested_remove_attachments(self):
        """A forwarded mail with attachments results in nested multipart
        payloads - this should also be handled by remove_attachments.
        """
        msg = deepcopy(self.msg_nested_attachments)
        self.assertNotEquals(msg, self.msg_nested_attachments)

        # we have two attachments (which are nested)
        self.assertEquals([{'position': 4,
                            'size': 137588,
                            'content-type': 'image/jpg',
                            'filename': '1703693_0412c29a4f.jpg'},
                           {'position': 5,
                            'size': 223504,
                            'content-type': 'image/jpg',
                            'filename': '3512536451_e1310bf568.jpg'}],
                          utils.get_attachments(msg))

        # lets remove one attachment
        new_msg = utils.remove_attachments(msg, (5,))

        # we get the same message back..
        self.assertEquals(msg, new_msg)
        # .. but without the removed attachment
        self.assertEquals([{'position': 4,
                            'size': 137588,
                            'content-type': 'image/jpg',
                            'filename': '1703693_0412c29a4f.jpg'}],
                          utils.get_attachments(new_msg))

    def test_unwrap_html_body(self):
        html = """
        <html>
        <head></head>
        <body>Body</body>
        </html>
        """
        body = '<div>Body</div>'
        self.assertEquals(body, utils.unwrap_html_body(html))
        html = """
        <html>
        <body style="color: #666; font-size: 12px;">Body</body>
        </html>
        """
        body = '<div class="mailBody" style="color: #666; font-size: 12px;">Body</div>'
        self.assertEquals(body, utils.unwrap_html_body(html, 'mailBody'))
        html = '<p>Body</p>'
        body = '<div class="mailBody"><p>Body</p></div>'
        self.assertEquals(body, utils.unwrap_html_body(html, 'mailBody'))

    def test_unwrap_html_body_encoding(self):
        # the html body may contain a charset header
        # we always get an utf8-encoded body, thus we must ignore the charset
        html = """
        <html>
        <head>
        <meta http-equiv=Content-Type content="text/html; charset=iso-8859-1">
        </head>
        <body>Äöü</body>
        """
        self.assertEquals('<div>Äöü</div>', utils.unwrap_html_body(html))

        # BeautifulSoup does some weird encoding guessing.
        # For the snippet above it guesses utf-8, but if the body
        # only contains a single ä Umlaut, it seems to guess latin1.
        # Check we still get utf-8 back.
        html = """
        <html>
        <body>ä</body>
        """
        self.assertEquals('<div>ä</div>', utils.unwrap_html_body(html))

    def test_unwrap_attached_msg(self):
        msg = utils.unwrap_attached_msg(self.msg_fwd_attachment)
        self.assertEquals(msg.get('Subject'), 'Lorem Ipsum')

    def test_get_position_for_cid(self):
        self.assertEqual(4, utils.get_position_for_cid(
            self.nested_referenced_image_attachment,
            'BB5DB00F-5C9A-4866-894D-8468D4B320F8'))

    def test_adjust_image_tags_generates_correct_links(self):
        html = ''.join(
            utils.get_text_payloads(self.nested_referenced_image_attachment))

        self.assertNotIn('get_attachment?position=4', html)
        html = utils.adjust_image_tags(html,
                                       self.nested_referenced_image_attachment,
                                       'foo')
        self.assertIn('get_attachment?position=4', html)


def test_suite():
    return unittest2.defaultTestLoader.loadTestsFromName(__name__)



from ftw.mail.mail import IMail
from plone.indexer.decorator import indexer


@indexer(IMail)
def Date(object_, **kw):
    """indexes the date of emails"""
    email_date = object_.get_header("Date", True) or object_.created()
    return email_date



from ftw.mail.interfaces import IEmailAddress
from ftw.mail.interfaces import IMailSettings
from plone.app.uuid.utils import uuidToObject
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from zope.component import queryUtility
from zope.interface import implements


class UUIDEmailAddress(object):
    implements(IEmailAddress)

    def __init__(self, request):
        self.request = request

    def get_email_for_object(self, obj, domain=None):
        if domain is None:
            registry = queryUtility(IRegistry)
            proxy = registry.forInterface(IMailSettings)
            domain = getattr(proxy, 'mail_domain', u'nodomain.com').encode(
            'utf-8')

        return '@'.join((IUUID(obj), domain))

    def get_object_for_email(self, email):
        uuid = email.split('@')[0]
        return uuidToObject(uuid)



from Acquisition import aq_inner
from ftw.mail import utils
from ftw.mail.utils import walk
from Products.Five.browser import BrowserView
from zExceptions import NotFound
import email


class AttachmentView(BrowserView):
    """Returns the attachment at the position specified in the request.
    """

    def __call__(self):
        context = aq_inner(self.context)
        self.message = context.message
        self.position = self.request.get('position', '0')

        return self.render()

    def render(self):

        if self.message is None:
            raise NotFound
        message = email.message_from_string(self.message.data)

        # we need an int value for the position
        pos = 0
        try:
            pos = int(self.position)
        except ValueError:
            raise NotFound

        # get attachment at position pos
        attachment = None
        for i, part in enumerate(walk(message)):
            if i == pos:
                attachment = part
                continue

        if attachment is not None:
            content_type = attachment.get_content_type()
            filename = utils.get_filename(attachment)
            if filename is None:
                raise NotFound
            # make sure we have a unicode string
            if not isinstance(filename, unicode):
                filename = filename.decode('utf-8', 'ignore')
            # internet explorer and safari don't like rfc encoding of filenames
            # and they don't like utf encoding too.
            # therefore we first try to encode the filename in iso-8859-1
            try:
                filename = filename.encode('iso-8859-1')
            except:
                filename = filename.encode('utf-8', 'ignore')

            self.request.response.setHeader('Content-Type', content_type)
            self.request.response.setHeader('Content-Disposition', 'inline; filename=%s' % filename)

            if content_type == 'message/rfc822':
                return attachment.as_string()
            return attachment.get_payload(decode=1)

        raise NotFound



from ftw.builder import builder_registry
from ftw.builder.dexterity import DexterityBuilder
from plone.namedfile import NamedBlobFile


class MailBuilder(DexterityBuilder):
    portal_type = 'ftw.mail.mail'

    def with_message(self, data, contentType=u'message/rfc822', filename=u'message.eml'):
        if isinstance(data, file):
            data = data.read()
        self.arguments["message"] = NamedBlobFile(
            data=data, contentType=contentType, filename=filename)
        return self


builder_registry.register('mail', MailBuilder)



from AccessControl import getSecurityManager
from AccessControl import Unauthorized
from Acquisition import aq_base
from Acquisition import aq_inner
from ftw.mail.interfaces import ICreateMailInContainer
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.utils import createContent
from plone.dexterity.utils import iterSchemata
from plone.i18n.normalizer.interfaces import IIDNormalizer
from z3c.form.interfaces import IValue
from zope.component import getUtility
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.container.interfaces import INameChooser
from zope.interface import implements
from zope.schema import getFields
from zope.schema import getFieldsInOrder
from zope.security.interfaces import IPermission


class CreateMailInContainer(object):
    implements(ICreateMailInContainer)

    portal_type = 'ftw.mail.mail'

    def __init__(self, context, request):
        self.context = aq_inner(context)
        self.request = request

    def create_mail(self, message):
        """Add a mail object to a container.

        The new object, wrapped in its new acquisition context, is returned.
        """

        self.check_permission()
        self.check_addable_types()

        content = self.create_mail_object(message)
        name = self.choose_name(content)
        obj = self.add_to_container(name, content)

        self.set_defaults(obj)
        return obj

    def check_permission(self):
        permission = queryUtility(IPermission, name='ftw.mail.AddInboundMail')
        if permission is None:
            raise Unauthorized("Cannot create %s" % self.portal_type)
        if not getSecurityManager().checkPermission(permission.title,
                                                    self.context):
            raise Unauthorized("Cannot create %s" % self.portal_type)

    def check_addable_types(self):
        container_fti = self.context.getTypeInfo()
        if container_fti is not None and \
                not container_fti.allowType(self.portal_type):
            raise ValueError("Disallowed subobject type: %s" % (
                self.portal_type))

    def create_mail_object(self, message):
        # lookup the type of the 'message' field and create an instance
        fti = getUtility(IDexterityFTI, name=self.portal_type)
        schema = fti.lookupSchema()
        field_type = getFields(schema)['message']._type
        message_value = field_type(
            data=message,
            contentType='message/rfc822', filename=u'message.eml')

        # create content object
        content = createContent(self.portal_type, message=message_value)
        return content

    def choose_name(self, content):
        normalizer = queryUtility(IIDNormalizer)
        normalized_subject = normalizer.normalize(content.title)

        name = INameChooser(self.context).chooseName(
            normalized_subject, content)
        content.id = name
        return name

    def add_to_container(self, name, content):
        newName = self.context._setObject(name, content)
        obj = self.context._getOb(newName)
        return obj

    def set_defaults(self, obj):
        """set the default value for all fields on the mail object
        (including additional behaviors)"""

        for schema in iterSchemata(obj):
            for name, field in getFieldsInOrder(schema):
                # Remove acquisition wrapper when getting field value so
                # determining if a field is already set works as expected
                value = field.get(field.interface(aq_base(obj)))
                if value == field.missing_value or value is None:
                    # bind the field for choices with named vocabularies
                    field = field.bind(obj)

                    # No value is set, so we try to set the default value
                    # otherwise we set the missing value
                    default = queryMultiAdapter((
                            self.context,
                            self.request,  # request
                            None,  # form
                            field,
                            None,  # Widget
                            ), IValue, name='default')
                    if default is not None:
                        default = default.get()
                    if default is None:
                        default = getattr(field, 'default', None)
                    if default is None:
                        try:
                            default = field.missing_value
                        except:
                            pass
                    field.set(field.interface(obj), default)

        obj.reindexObject()



from setuptools import find_packages
from setuptools import setup
import os


version = '2.7.0'

tests_require = [
    'ftw.builder',
    'ftw.testbrowser',
    'ftw.workspace',
    'ftw.zipexport',
    'plone.app.testing',
    'unittest2',
    ]

setup(name='ftw.mail',
      version=version,
      description='Provides a mail content type and a mail-in behavior',
      long_description=open('README.rst').read() + '\n' + \
          open(os.path.join('docs', 'HISTORY.txt')).read(),

      # Get more strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.3',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],

      keywords='',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.mail',
      license='GPL2',

      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
        'plone.api',
        'plone.app.referenceablebehavior',
        'plone.registry',
        'plone.dexterity',
        'setuptools',
        'BeautifulSoup!=4.0b',
        'plone.app.dexterity',
        'plone.app.registry',
        'plone.namedfile[blobs]',
        'collective.autopermission',
        'collective.dexteritytextindexer',
        'plone.directives.form',
        'ftw.upgrade >= 1.11.0',
        'premailer < 3.7.0',
        ],

      tests_require=tests_require,
      extras_require=dict(tests=tests_require),
      entry_points='''
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      ''',
      )



from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from ftw.mail.tests import builders  # noqa
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.testing import z2
from zope.configuration import xmlconfig


class FtwMailLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE, BUILDER_LAYER)

    def setUpZope(self, app, configurationContext):
        xmlconfig.string(
            '<configure xmlns="http://namespaces.zope.org/zope">'
            '  <include package="z3c.autoinclude" file="meta.zcml" />'
            '  <includePlugins package="plone" />'
            '  <includePluginsOverrides package="plone" />'
            '</configure>',
            context=configurationContext,
        )

        import Products.CMFPlacefulWorkflow
        xmlconfig.file('configure.zcml', Products.CMFPlacefulWorkflow, context=configurationContext)

        try:
            # Plone >= 4.3.x
            import Products.DataGridField
            xmlconfig.file('configure.zcml', Products.DataGridField, context=configurationContext)
        except:
            pass

        z2.installProduct(app, 'ftw.file')
        z2.installProduct(app, 'ftw.meeting')
        z2.installProduct(app, 'ftw.workspace')

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.mail:default')
        applyProfile(portal, 'ftw.workspace:default')
        applyProfile(portal, 'ftw.zipexport:default')


FTW_MAIL_LAYER = FtwMailLayer()
FTW_MAIL_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(FTW_MAIL_LAYER, set_builder_session_factory(functional_session_factory)),
    name="ftw.mail:functional",
)



from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from ftw.builder import Builder
from ftw.builder import create
from ftw.mail.testing import FTW_MAIL_FUNCTIONAL_TESTING
from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles
from plone.uuid.interfaces import IUUID
from unittest2 import TestCase
from zope.component import queryMultiAdapter
from zope.viewlet.interfaces import IViewletManager


class TestMailInViewlet(TestCase):

    layer = FTW_MAIL_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer.get('portal')
        setRoles(self.portal, TEST_USER_ID, ['Manager', 'Member'])

        self.folder = create(Builder('folder'))

    def get_viewlet(self, context):
        view = BrowserView(context, context.REQUEST)
        manager_name = 'plone.belowcontenttitle'
        manager = queryMultiAdapter(
            (context, context.REQUEST, view),
            IViewletManager,
            manager_name)

        self.failUnless(manager)

        # Set up viewlets
        manager.update()
        name = 'ftw.mail.mail-in'
        viewlets = [v for v in manager.viewlets if v.__name__ == name]

        if len(viewlets):
            return viewlets[0]
        else:
            return None

    def test_viewlet_registered(self):
        self.assertNotEquals(None, self.get_viewlet(self.folder),
            'The ftw.mail.mail-in viewlet is not registered properly')

    def test_viewlet_is_present_if_mail_is_addable(self):
        portal_types = getToolByName(self.portal, 'portal_types')
        portal_types.get('Folder').allowed_content_types = ['ftw.mail.mail']

        viewlet = self.get_viewlet(self.folder)

        self.assertTrue(viewlet.available(),
            'Expect the ftw.mail-mail-in viewlet on %s' % (
                self.folder.absolute_url()))

    def test_viewlet_is_not_present_if_mail_is_not_addable(self):
        portal_types = getToolByName(self.portal, 'portal_types')
        portal_types.get('Folder').allowed_content_types = []

        viewlet = self.get_viewlet(self.folder)

        self.assertFalse(viewlet.available(),
            'Did not expect the ftw.mail-mail-in viewlet on %s' % (
                self.folder.absolute_url()))

    def test_mailin_email(self):
        viewlet = self.get_viewlet(self.folder)
        expect = "%s@example.org" % IUUID(self.folder)

        self.assertEquals(expect, viewlet.email())
